<?php

namespace Drupal\migrate_source_graphql\Plugin\migrate\source;

use Drupal\Component\Plugin\ConfigurableInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\migrate\MigrateMessage;
use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\migrate\Plugin\migrate\source\SourcePluginBase;
use Drupal\migrate\Row;
use Drupal\migrate_source_graphql\Event\ResultsEvent;
use Drupal\migrate_source_graphql\GraphQL\Client;
use GraphQL\Exception\QueryError;
use GraphQL\Query;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class GraphQL migrate source.
 *
 * @MigrateSource(
 *   id = "graphql",
 *   source_module = "migrate_source_graphql"
 * )
 */
class GraphQL extends SourcePluginBase implements ConfigurableInterface, ContainerFactoryPluginInterface {
  const DATA_KEY_INDEXES = '%';

  /**
   * The graphql client.
   *
   * @var \GraphQL\Client
   */
  protected $client;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Description of the unique ID fields for this source.
   *
   * @var array[]
   *   Each array member is keyed by a field name, with a value that is an
   *   array with a single member with key 'type' and value a column type such
   *   as 'integer'.
   */
  protected $ids = [];

  /**
   * {@inheritdoc}
   *
   * @throws \InvalidArgumentException
   * @throws \Drupal\migrate\MigrateException
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, MigrationInterface $migration) {
    $configuration['data_key'] = isset($configuration['data_key']) ? $configuration['data_key'] : 'data';
    $this->ids = array_key_exists('ids', $configuration) ? $configuration['ids'] : ['id' => ['type' => 'string']];

    parent::__construct($configuration, $plugin_id, $plugin_definition, $migration);
    // Endpoint is required.
    if (empty($this->configuration['endpoint'])) {
      throw new \InvalidArgumentException('You must declare the "endpoint" to the GraphQL API service in your settings.');
    }

    // Queries are required.
    if (empty($this->configuration['query'])) {
      throw new \InvalidArgumentException('You must declare the "query" parameter in your settings to get expected data from GraphQL API service.');
    }
    else {
      $headers = [];
      if (isset($this->configuration['auth_scheme']) && !empty($this->configuration['auth_scheme'])) {
        $headers['Authorization'] = $this->configuration['auth_scheme'] . ' ' . ($this->configuration['auth_parameters'] ?? '');
      }
      $this->client = new Client($this->configuration['endpoint'], $headers);
      $this->fields();
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition, MigrationInterface $migration = NULL): GraphQL {
    $instance = new static($configuration, $plugin_id, $plugin_definition, $migration);
    $instance->eventDispatcher = $container->get('event_dispatcher');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'endpoint' => 'localhost',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    // We must preserve integer keys for column_name mapping.
    $this->configuration = NestedArray::mergeDeepArray([
      $this->defaultConfiguration(),
      $configuration,
    ], TRUE);
  }

  /**
   * Return a string representing the GraphQL API endpoint.
   *
   * @return string
   *   The GraphQL API endpoint.
   */
  public function __toString() {
    return $this->configuration['endpoint'];
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\migrate\MigrateException
   */
  public function initializeIterator() {
    return $this->getGenerator();
  }

  /**
   * Return the generator using yield.
   */
  private function getGenerator() {
    $query = $this->configuration['query'];
    $queryName = array_key_first($query);
    $this->client->buildQueryRecursive($queryName, $query[$queryName], $qlQuery);
    try {
      $results  = $this->client->runQuery($qlQuery);
      $results  = $results->getData()->$queryName ?? [];
      $propertyPath = explode(Row::PROPERTY_SEPARATOR, $this->configuration['data_key']);
      $getIndexes = false;
      foreach ($propertyPath as $path) {
        if ($path === self::DATA_KEY_INDEXES) {
          $getIndexes = true;
          continue;
        }
        if ($getIndexes) {
          $results = array_map(function ($item) use ($path) {
            return $item->{$path};
          }, array_values($results));
          $getIndexes = false;
        } else {
          $results = $results->{$path} ?? [];
        }
      }
      // Allow other modules to manipulate the results.
      $resultsEvent = new ResultsEvent($this->migration, new MigrateMessage(), $results);
      $this->eventDispatcher->dispatch($resultsEvent);
      $results = $resultsEvent->getResults();
      foreach ($results as $result) {
        yield json_decode(json_encode($result), TRUE);
        ;
      }
    } catch (QueryError $exception) {
      \Drupal::messenger()->addError($exception->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getIds() {
    return $this->ids;
  }

  /**
   * {@inheritdoc}
   */
  public function fields() {
    $fields = [];
    $query = $this->configuration['query'];
    foreach ($query as $queryName => $query) {
      $results = is_array($query['fields'][0]) ? array_keys($query['fields'][0]) : $query['fields'];
      foreach ($results as $resultKey) {
        if (isset($query['fields'][0][$resultKey])) {
          foreach ($query['fields'][0][$resultKey] as $field) {
            if (!is_array($field)) {
              $fields[$field] = $field;
            }
          }
        }
        else {
          $fields[$resultKey] = $resultKey;
        }
      }
    }
    return $fields;
  }
}
